#include "async.h"
#include <sys/types.h>
#include <sys/socket.h>

#define FD_MAX 64
#define NCON_MAX FD_MAX - 8
#define FINGER_PORT 79

int ncon;
int argno, gargc;
char **gargv;

void finger(char * arg);
void connect_cb(str host, str user, int fd);
void write_cb(str user, int fd, int written, int len);
void read_cb(int, char *, int, int);
void done(int fd, char *resp_buf);

int main(int argc, char **argv) {
  ncon = 0; argno = 1;
  gargv = argv; gargc = argc;
  while ( (argno < argc) && (ncon < NCON_MAX)) finger(argv[argno++]);
  amain();
}

void
finger(char * arg) {
  char *p = strrchr(arg, '@');
  if (!p) {
    fprintf (stderr, "%s: ignored -- not of form 'user@host'\n", arg);
    return;
  }
  *p++ = 0;

  //str is a refcounted type no need to worry about freeing its memory
  str host (p);
  str user = strbuf() << arg << "\n\r";
  ncon++;
  tcpconnect(host.cstr(), FINGER_PORT, wrap(&connect_cb, host, user));
}

void
connect_cb(str host, str user, int fd) {
  if (fd < 0) {
    warn << "unable to connect to " << host << "(" << strerror(errno) << ")\n";
    done(-1, NULL);
    return;
  }  
  write_cb(user, fd, 0, user.len());
}

void
write_cb(str user, int fd, int written, int len) {
  int err = write(fd, user.cstr() + written, len - written);
  if ((err == -1) && (errno != EAGAIN)) {
    warn << "error writing to socket" << user << "\n";
    done(fd, NULL);
    return;
  } 
  int bytesWritten = ((err == -1) && (errno == EAGAIN)) ? 0 : err;
  if (bytesWritten < len) 
    fdcb(fd, selwrite, wrap(&write_cb, user, fd, bytesWritten + written, len - bytesWritten));
  else {
    fdcb(fd, selwrite, NULL);
    char *resp_buf = new char[2048];
    fdcb(fd, selread, wrap(&read_cb, fd, resp_buf, 0, 2048));
  }
}

void
read_cb(int fd, char *resp_buf, int offset, int buf_size) {

  if (512 > buf_size - offset)
    resp_buf = (char *)realloc((void *)resp_buf, (buf_size *= 2));
  int err = read(fd, resp_buf + offset, buf_size - offset);
  if (err == -1) {
    warn << "couldn't read from socket: " << strerror(errno) << "\n";
    done(fd, resp_buf);
    return;
  } else if (err == 0) {
    write(1, resp_buf, offset + err);
    done(fd, resp_buf);
    return;
  } 
  fdcb(fd, selread, wrap(&read_cb, fd, resp_buf, offset + err, buf_size));
  return;
}

void
done(int fd, char *resp_buf) {
  if (fd > 0) {
    close(fd);
    fdcb(fd, selread, NULL);
    fdcb(fd, selwrite, NULL);
  }
  if (resp_buf) free(resp_buf);
  if ((--ncon == 0) && (argno == gargc)) exit(0);
  //queue up more transactions
  while ((argno < gargc) && (ncon < NCON_MAX)) finger(gargv[argno++]);
}
